from thompcoutils.log_utils import get_logger
from thompcoutils.config_utils import ConfigManager
import logging
import sqlobject
import os


class DBConfig:
    def __init__(self, cfg_mgr=None, db_type=None, sqlite_file=None, host=None, username=None, password=None,
                 schema=None, port=None, rebuild=None, section="db connection"):
        if cfg_mgr is None:
            self.type = db_type
            self.sqlite_file = sqlite_file
            self.host = host
            self.username = username
            self.password = password
            self.schema = schema
            self.port = port
            self.rebuild = rebuild
        else:
            self.type = cfg_mgr.read_entry(section, "type", default_value="mysql")
            self.sqlite_file = cfg_mgr.read_entry(section, "sqlite file", default_value="licenseserver.sqlite")
            self.host = cfg_mgr.read_entry(section, "host", default_value="localhost")
            self.username = cfg_mgr.read_entry(section, "username", default_value="my db user name")
            self.password = cfg_mgr.read_entry(section, "password", default_value="my db user password")
            self.schema = cfg_mgr.read_entry(section, "schema", default_value="my db schema")
            self.port = cfg_mgr.read_entry(section, "port", default_value=3306)
            self.rebuild = cfg_mgr.read_entry(section, "rebuild", default_value=False)


class DbUtils:
    def __init__(self, db_cfg=None, db_type=None, host=None, username=None, password=None, schema=None, port=None,
                 sqlite_file=None):
        self.connection = None
        if db_cfg is not None:
            db_type = db_cfg.type
            host = db_cfg.host
            username = db_cfg.username
            password = db_cfg.password
            schema = db_cfg.schema
            port = db_cfg.port
        if db_type == "sqlite":
            if sqlite_file is None:
                raise RuntimeError("--sqlite requires --sqlite-dir")
            else:
                print("Using Sqlite database")
                self._connect_sqlite(sqlite_file)
        elif db_type == "postgres":
            self._connect_postgres(username, password, schema, host, port)
        elif "mysql" in db_type:
            self._connect_mysql(username, password, schema, host, port)
        elif db_type == "odbc":
            raise RuntimeError("ODBC not implemented yet")
        else:
            raise RuntimeError("No database type selected")

    def _connect_uri(self, uri):
        self.connection = sqlobject.sqlhub.processConnection = sqlobject.connectionForURI(uri)

    def _connect_sqlite(self, file_path):
        uri = "sqlite:" + file_path
        self._connect_uri(uri)

    def _connect_postgres(self, username, password, database, host, port):
        port_str = ""
        if port is not None:
            port_str = ":" + str(port)
        uri = "postgres://" + username + ":" + password + "@" + host + port_str + "/" + database
        self._connect_uri(uri)

    def _connect_mysql(self, username, password, database, host, port):
        port_str = ""
        if port is not None:
            port_str = ":" + str(port)
        uri = "mysql://" + username + ":" + password + "@" + host + port_str + "/" + database
        self._connect_uri(uri)

    def create_table(self, table):
        logger = get_logger()
        if not self.connection.tableExists(table.q.tableName):
            logger.debug("creating table {}".format(str(table)))
            table.createTable()

    def create_tables(self, tables):
        logger = get_logger()
        logger.debug("Creating tables...")
        for table in tables:
            self.create_table(table)

    @staticmethod
    def drop_tables(tables):
        logger = get_logger()
        logger.debug("Dropping tables...")
        for table in tables:
            logger.debug("dropping table {}".format(str(table)))
            table.dropTable(cascade=True, ifExists=True)


def main():
    create_file = False
    logging.config.fileConfig('logging.ini')
    temp_filename = "testing DbConfig only.cfg"
    if create_file:
        if os.path.isfile(temp_filename):
            os.remove(temp_filename)
    cfg_mgr = ConfigManager(temp_filename, create=create_file)
    db_cfg = DBConfig(cfg_mgr=cfg_mgr)
    if create_file:
        cfg_mgr.write(temp_filename)
    db = DbUtils(db_cfg)


if __name__ == "__main__":
    main()



import abc
import copy
import os
import pathlib

SPACER = '  '


class HtmlItem(object):
    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def get_style(self, depth):
        pass

    @abc.abstractmethod
    def get_body(self, depth):
        pass


class Html:
    def __init__(self):
        self.__html_items = []

    def to_string(self):
        rtn = '<html>\n'
        rtn += '{}<style>\n'.format(SPACER)
        for html_item in self.__html_items:
            style = html_item.get_style(depth=2)
            if style.strip() != '':
                rtn += html_item.get_style(depth=2) + '\n'
        rtn += '{}</style>\n'.format(SPACER)
        for html_item in self.__html_items:
            rtn += html_item.get_body(depth=1) + '\n'
        rtn += '</html>\n'
        return rtn

    def append(self, html_item):
        if issubclass(html_item.__class__, HtmlItem):
            self.__html_items.append(html_item)
        else:
            raise Exception('object is not a HtmlItem')


class HtmlFooter(HtmlItem):
    def __init__(self, html_items):
        for html_item in html_items:
            if not issubclass(html_item.__class__, HtmlItem):
                raise Exception('object is not a HtmlItem')
        self.html_items = html_items

    def get_style(self, depth):
        rtn = ''
        for html_item in self.html_items:
            rtn += html_item.get_body() + '\n'
        return rtn

    def get_body(self, depth):
        rtn = '<footer>\n'
        for html_item in self.html_items:
            rtn += html_item.get_body() + '\n'
        rtn += '</footer>\n'
        return rtn


class HtmlList(HtmlItem):
    def __init__(self, html_items, ordered=True):
        for html_item in html_items:
            if not issubclass(html_item.__class__, HtmlItem):
                raise Exception('object is not a HtmlItem')
        self.html_items = html_items
        self.ordered = ordered

    def get_style(self, depth):
        rtn = ''
        for item in self.html_items:
            rtn += item.get_style(depth)
        return rtn

    def get_body(self, depth):
        rtn = ''
        for item in self.html_items:
            rtn += '{}<li>{}</li>\n'.format(SPACER * (depth + 1), item.get_body(0))
        if self.ordered:
            rtn = '{}<ol>\n{}{}</ol>'.format(SPACER * depth, rtn, SPACER * depth)
        else:
            rtn = '{}<ul>\n{}{}</ul>'.format(SPACER * depth, rtn, SPACER * depth)
        return rtn


class HtmlCode(HtmlItem):
    def __init__(self, text):
        self.text = text

    def get_style(self, depth):
        return ''

    def get_body(self, depth):
        text = self.text.replace('\n', '<br>\n' + SPACER * (depth+1))
        return '{}<code>\n{}{}\n{}</code>'.format(SPACER * depth, SPACER * (depth + 1), text, SPACER * depth)


class HtmlText(HtmlItem):
    def __init__(self, text, bold=False, italicized=False, centered=False, heading_level=None,
                 paragraph=None, font=None):
        self.text = text
        self.bold = bold
        self.italicized = italicized
        self.centered = centered
        self.heading_level = heading_level
        self.paragraph = paragraph
        self.font = font

    def get_style(self, depth):
        return ''

    def get_body(self, depth):
        text = self.text.replace('\n', '<br>\n' + SPACER * (depth+1))
        if self.font is None:
            rtn = text
        else:
            rtn = self.font.to_string(text)
        if self.bold:
            rtn = '<b>{}</b>'.format(rtn)
        if self.italicized:
            rtn = '<i>{}</i>'.format(rtn)
        if self.centered:
            rtn = '<c>{}</c>'.format(rtn)
        if self.heading_level:
            rtn = '<h{}>{}</h{}>'.format(self.heading_level, rtn, self.heading_level)
        if self.paragraph:
            rtn = '<p>{}</p>'.format(rtn)
        return SPACER * (depth + 1) + rtn


class HtmlTable(HtmlItem):
    def __init__(self, data, column_headings=None, border=None, border_collapse=None,
                 font_family=None, width=None, text_align=None, padding=None,
                 even_background_color=None,
                 odd_background_color=None):
        self.data = data
        self.column_headings = column_headings
        self.border_collapse = border_collapse
        self.border = border
        self.font_family = font_family
        self.width = width
        self.text_align = text_align
        self.padding = padding
        self.even_background_color = even_background_color
        self.odd_background_color = odd_background_color

    def add_row(self, row):
        self.data.append(row)

    def get_style(self, depth):
        rtn = ''
        rtn += '{}table {{\n'.format(SPACER * depth)
        if self.font_family is not None:
            rtn += '{}font-family: {};\n'.format(SPACER * (depth + 1), self.font_family)
        if self.border_collapse is not None:
            rtn += '{}border-collapse: {};\n'.format(SPACER * (depth + 1), self.border_collapse)
        if self.width is not None:
            rtn += '{}width: {};\n'.format(SPACER * (depth + 1), self.width)
        rtn += '{}}}\n'.format(SPACER * depth)
        rtn += '{}td, th {{\n'.format(SPACER * depth)
        if self.border is not None:
            rtn += '{}border: {};\n'.format(SPACER * (depth + 1), self.border)
        if self.text_align is not None:
            rtn += '{}text-align: {};\n'.format(SPACER * (depth + 1), self.text_align)
        if self.padding is not None:
            rtn += '{}padding: {}\n'.format(SPACER * (depth + 1), self.padding)
        rtn += '{}}}'.format(SPACER * depth)
        return rtn

    def get_body(self, depth):
        rtn = ''
        rtn += '{}<table>\n'.format(SPACER * depth)
        if self.column_headings is not None:
            rtn += '{}<tr>\n'.format(SPACER * (depth + 1))
            for heading in self.column_headings:
                rtn += '{}<th>{}</th>\n'.format(SPACER * (depth + 2), heading)
            rtn += '{}</tr>\n'.format(SPACER * (depth + 1))
        line_number = 0
        for row in self.data:
            rtn += '{}<tr>\n'.format(SPACER * (depth + 1))
            for cell in row:
                bg_color = ''
                if line_number % 2:
                    if self.even_background_color:
                        bg_color = ' bgcolor={}'.format(self.even_background_color)
                else:
                    if self.odd_background_color:
                        bg_color = ' bgcolor={}'.format(self.odd_background_color)
                rtn += '{}<td{}>{}</td>\n'.format(SPACER * (depth + 2), bg_color, cell)
            rtn += '{}</tr>\n'.format(SPACER * (depth + 1))
            line_number += 1
        rtn += '{}</table>'.format(SPACER * depth)
        return rtn


class HtmlBreak(HtmlItem):
    def __init__(self):
        pass

    def get_style(self, depth):
        return ''

    def get_body(self, depth):
        return '{}<br>'.format(SPACER * depth)


class HtmlLine(HtmlItem):
    def __init__(self):
        pass

    def get_style(self, depth):
        return ''

    def get_body(self, depth):
        return '<hr>'


class HtmlImage(HtmlItem):
    def __init__(self, image, alt_text=None, height=None, width=None):
        self.image = image
        self.alt_text = alt_text
        self.height = height
        self.width = width

    def get_style(self, depth):
        return ''

    def get_body(self, depth):
        if self.alt_text:
            alt_text = ' alt="{}"'.format(self.alt_text)
        else:
            alt_text = ''
        if self.height:
            height = ' height="{}"'.format(self.height)
        else:
            height = ''
        if self.width:
            width = ' width="{}"'.format(self.width)
        else:
            width = ''
        return '<img src="{}"{}{}{}>'.format(self.image, alt_text, height, width)


class HtmlLink(HtmlItem):
    def __init__(self, link, text):
        self.link = link
        self.text = text

    def get_style(self, depth):
        return ''

    def get_body(self, depth):
        return '{}<a href="{}">{}</a>'.format(SPACER * depth, self.link, self.text)


class HtmlMailTo(HtmlLink):
    def __init__(self, email, user_name):
        super(HtmlLink, self).__init__()
        self.link = 'mailto:{}'.format(email)
        self.text = user_name


class Font:
    def __init__(self, color=None, size=None, face=None):
        self.color = color
        self.size = size
        self.face = face

    def to_string(self, text):
        rtn = '<font'
        if self.size is not None:
            rtn += ' size="{}"'.format(self.size)
        if self.color is not None:
            rtn += ' color="{}"'.format(self.color)
        if self.face is not None:
            rtn += ' face="{}"'.format(self.face)
        rtn += '>{}</font>'.format(text)
        return rtn


def main():
    file_name = 'test.html'
    with open(file_name, 'w') as the_file:
        html = Html()
        values = [HtmlText('first', bold=True), HtmlText('second'), HtmlText('third')]
        ol = HtmlList(values)
        html.append(ol)
        ul = HtmlList(values, ordered=False)
        html.append(ul)
        text = HtmlText('this is some text\nwith a carriage return\nin several places')
        red = Font(color='red')
        text.font = red
        red.size = 1
        red.face = 'verdana'
        ol.html_items.append(copy.deepcopy(text))
        html.append(ol)
        text.paragraph = True
        text.font.color = 'blue'
        html.append(copy.deepcopy(text))
        text.font.size = None
        text.heading_level = 5
        html.append(copy.deepcopy(text))
        text.font.color = 'green'
        html.append(copy.deepcopy(text))
        headings = ['AAAA', 'BBBB', 'CCCC']
        data = [['a1', 'b1', 'c1'],
                ['a2', 'b2', 'c2'],
                ['a3', 'b3', 'c3'],
                ['a4', 'b4', 'c4'],
                ['a5', 'b5', 'c5']]
        table = HtmlTable(data=data, column_headings=headings)
        table.border = '1px solid #dddddd'
        table.text_align = 'center'
        #        table.padding = '8px'
        table.font_family = 'arial, sans-serif'
        table.border_collapse = 'collapse'
        table.width = '1000px'
        table.even_background_color = '#dddddd'
        table.odd_background_color = '#dddd00'
        html.append(table)
        image_file = os.path.join(os.path.join(pathlib.Path().parent.absolute().parent, 'attachments'), 'image_1.png')
        image = HtmlImage(image_file, height=100, width=120, alt_text='testing')
        html.append(HtmlLine())
        html.append(image)
        html.append(HtmlBreak())
        html.append(HtmlMailTo('jordan@thompco.com', 'Jordan Thompson'))
        html.append(HtmlBreak())
        html.append(HtmlCode('This is a test\nAnd another\nand another\nand yet another!'))
        the_file.write(html.to_string())
        the_file.close()
        os.system('python3 -m webbrowser -t "file://{}"'.format(os.path.join(pathlib.Path().absolute(), file_name)))


if __name__ == '__main__':
    main()



import os
from thompcoutils.log_utils import get_logger
import psutil
from sys import platform
import time
from netifaces import interfaces, ifaddresses, AF_INET


class UnhandledOs(Exception):
    pass


def _list_zones(walk_dir):
    d = {"timezones": []}
    files = os.listdir(walk_dir)
    for f in files:
        if os.path.isdir(os.path.join(walk_dir, f)):
            d[f] = _list_zones(os.path.join(walk_dir, f))
        else:
            d["timezones"].append(f)
    return d


def list_timezones():
    ost = os_type()
    if ost == "windows":
        time_zones = _list_zones("d:/temp/zoneinfo")
        time_zones["current_timezone"] = get_timezone()
        return time_zones
    elif ost == "linux" or ost == "osx":
        time_zones = _list_zones("/usr/share/zoneinfo")
        time_zones["current_timezone"] = get_timezone()
        return time_zones
    else:
        raise UnhandledOs("{} os not supported".format(ost))


def get_timezone():
    ost = os_type()
    if ost == "linux":
        with open("/etc/timezone") as f:
            return f.read().strip()
    elif ost == "windows" or ost == "osx":
        return time.tzname[0]
    else:
        raise UnhandledOs("{} os not supported".format(ost))


def list_all_processes():
    processes = []
    for process in psutil.process_iter():
        processes.append(process)
    return processes


def find_processes(process_name):
    processes = []
    for process in psutil.process_iter():
        if process_name == process.name():
            processes.append(process)
    return processes


def list_child_processes(pid):
    processes = []
    for process in psutil.process_iter():
        if process.ppid() == pid:
            processes.append(process)
    return processes


def kill_process(process):
    logger = get_logger()
    killed = False
    try:
        process.kill()
        if not process.is_running():
            logger.debug('Killed "{}" process'.format(process.name()))
            killed = True
    except Exception as e:
        logger.debug(e)
    if not killed:
        logger.debug("Could not kill pid {}/process name {}".format(process.pid, process.name()))


def kill_all_processes_by_name(process_name):
    logger = get_logger()
    killed = False
    for process in psutil.process_iter():
        if process_name in process.name():
            # noinspection PyBroadException
            try:
                print(process.ppid)
                process.kill()
                if not process.is_running():
                    logger.debug('Killed "{}" process'.format(process.name()))
                    killed = True
            except Exception as e:
                logger.debug(e)
    if not killed:
        logger.error('Could kill process name "{}"'.format(process_name))


def is_running(name):
    logger = get_logger()
    pid_list = psutil.pids()
    for pid in pid_list:
        # noinspection PyBroadException
        try:
            process = psutil.Process(pid)
            if name == process.name() and process.is_running():
                return True
        except Exception as e:
            logger.debug(e)
    return False


class OSType(enumerate):
    Linux = "Linux"
    OSX = "OSX"
    WINDOWS = "Windows"


def os_type():
    if platform == "linux" or platform == "linux2":
        return OSType.Linux
    elif platform == "darwin":
        return OSType.OSX
    elif platform == "win32":
        return OSType.WINDOWS
    else:
        raise UnhandledOs("Unknown system")


def get_ip_addresses():
    addresses = None
    for interface_name in interfaces():
        addresses = [i['addr'] for i in ifaddresses(interface_name).setdefault(AF_INET, [{'addr': 'No IP addr'}])]
    return addresses


def main():
    chrome_processes = find_processes("Google Chrome")
    for process in chrome_processes:
        child_processes = list_child_processes(process.pid)
        if len(child_processes) < 7:
            kill_process(process)
    pass


if __name__ == '__main__':
    main()



import threading
import time


def _test(name, value1, value2):
    print("thread {} working".format(name))
    time.sleep(1)
    value1 *= 5
    value2 *= 2
    return value1, value2


class MyThread (threading.Thread):
    THREAD_LOCK = threading.Lock()

    def __init__(self, function, *argv):
        threading.Thread.__init__(self)
        self.function = function
        self.args = argv
        self.rtn = None

    def run(self):
        MyThread.THREAD_LOCK.acquire()
        self.rtn = self.function(*self.args)
        MyThread.THREAD_LOCK.release()


def _main():
    thread1 = MyThread(_test, 'one', 1, 2)
    thread2 = MyThread(_test, 'two', 2, 3)
    thread1.start()
    thread2.start()
    print('main working')
    thread1.join()
    thread2.join()
    print(thread1.rtn[0], thread1.rtn[1])
    print(thread2.rtn)
    print ("Exiting Main Thread")

if __name__ == '__main__':
    _main()



from thompcoutils.log_utils import get_logger
from thompcoutils.debug_utils import assert_test
import thompcoutils.file_utils as file_utils
import datetime
import sys
from builtins import staticmethod
import ast
from configparser import ConfigParser, NoOptionError, NoSectionError, DuplicateSectionError
import os
import logging
from dateutil import parser


class ConfigException(Exception):
    pass


class EmailConnectionConfig:
    def __init__(self, cfg_mgr=None, username=None, password=None, from_user=None, smtp_host="smtp.gmail.com",
                 smtp_port=587,
                 application_name="Application Name", filename=None, create=None, section_heading="email connection",
                 username_tag="username", password_tag="password", from_tag="from", smtp_host_tag="smtp_host",
                 smtp_port_tag="port"):
        top_package = __import__(__name__.split('.')[0])

        if cfg_mgr is None:
            self.username = username
            self.password = password
            self.from_user = from_user
            self.smtp_host = smtp_host
            self.smtp_port = smtp_port
        else:
            self.username = cfg_mgr.read_entry(section_heading, username_tag, "myname@google.com")
            self.password = cfg_mgr.read_entry(section_heading, password_tag, "mySecretPassword")
            self.from_user = cfg_mgr.read_entry(section_heading, from_tag, application_name,
                                                "Where the email will come from")
            self.smtp_host = cfg_mgr.read_entry(section_heading, smtp_host_tag, smtp_host)
            self.smtp_port = cfg_mgr.read_entry(section_heading, smtp_port_tag, smtp_port)


class HiLow:
    """
    HiLow class provides a config file to store low and hi values.  If a low value for a tag/entry is lower than the
    existing value, it will be updated.  Similarly, if a hi value for is higher than the existing hi value, it will be
    updated.
    """
    smallest = -sys.maxsize-1
    biggest = sys.maxsize
    hi_tag = "hi"
    low_tag = "low"
    last_tag = "last"
    low_changed_tag = "low_changed"
    hi_changed_tag = "hi_changed"
    last_changed_tag = "last_changed"
    direction_tag = "direction"
    change_amount_tag = "change_amount"

    class Direction(enumerate):
        Up = "up"
        Down = "down"
        NoChange = "no change"

        @staticmethod
        def validate(string):
            if string == str(HiLow.Direction.Up):
                return HiLow.Direction.Up
            elif string == str(HiLow.Direction.Down):
                return HiLow.Direction.Down
            elif string == str(HiLow.Direction.NoChange):
                return HiLow.Direction.NoChange
            else:
                raise ConfigException("{} not recognized as a HiLow.Direction".format(string))

    def __init__(self, file_name):
        """
        :param file_name: name of the file to store values
        """
        self.file_name = file_name
        if not os.path.exists(file_name):
            file_utils.touch(file_name)
        self.cfg_mgr = ConfigManager(file_name)

    def read_values(self, entry):
        """
        gets the current values for a tag/entry
        :param entry: tag/entry`
        :return: a dictionary of values (i.e. {"hi": 10, "low": 2} )
        """
        try:
            hi = self.cfg_mgr.read_entry(section=entry, entry=self.hi_tag, default_value=self.smallest)
        except ValueError:
            hi = self.cfg_mgr.read_entry(section=entry, entry=self.hi_tag, default_value=self.smallest + .01)
        try:
            low = self.cfg_mgr.read_entry(section=entry, entry=self.low_tag, default_value=self.biggest)
        except ValueError:
            low = self.cfg_mgr.read_entry(section=entry, entry=self.low_tag, default_value=self.biggest - .01)
        hi_changed_time = self.cfg_mgr.read_entry(section=entry, entry=self.hi_changed_tag,
                                                  default_value=datetime.datetime.now())
        low_changed_time = self.cfg_mgr.read_entry(section=entry, entry=self.low_changed_tag,
                                                   default_value=datetime.datetime.now())

        direction = self.cfg_mgr.read_entry(section=entry, entry=self.direction_tag,
                                            default_value=str(self.Direction.NoChange))
        direction = self.Direction.validate(direction)

        try:
            last = self.cfg_mgr.read_entry(section=entry, entry=self.last_tag, default_value=0)
        except ValueError:
            last = self.cfg_mgr.read_entry(section=entry, entry=self.last_tag, default_value=0.01)

        try:
            changed_amount = self.cfg_mgr.read_entry(section=entry, entry=self.change_amount_tag, default_value=0)
        except ValueError:
            changed_amount = self.cfg_mgr.read_entry(section=entry, entry=self.change_amount_tag, default_value=0.01)
        last_changed = self.cfg_mgr.read_entry(section=entry, entry=self.last_changed_tag,
                                               default_value=datetime.datetime.now())
        return {self.hi_tag: hi, self.low_tag: low,
                self.low_changed_tag: low_changed_time, self.hi_changed_tag: hi_changed_time,
                self.last_tag: last, self.last_changed_tag: last_changed, self.direction_tag: direction,
                self.change_amount_tag: changed_amount}

    def write_value(self, entry, value):
        try:
            self.cfg_mgr.config.add_section(entry)
        except DuplicateSectionError:
            pass

        try:
            cfg_last = self.cfg_mgr.read_entry(section=entry, entry=self.last_tag, default_value=0)
        except ValueError:
            cfg_last = self.cfg_mgr.read_entry(section=entry, entry=self.last_tag, default_value=.01)
        if value < cfg_last:
            self.cfg_mgr.write_entry(section=entry, entry=self.change_amount_tag, value=cfg_last-value)
            self.cfg_mgr.write_entry(section=entry, entry=self.direction_tag, value=self.Direction.Down)
            try:
                cfg_low = self.cfg_mgr.read_entry(section=entry, entry=self.low_tag, default_value=self.biggest)
            except ValueError:
                cfg_low = self.cfg_mgr.read_entry(section=entry, entry=self.low_tag, default_value=self.biggest - .01)
            if value < cfg_low:
                self.cfg_mgr.write_entry(section=entry, entry=self.low_changed_tag, value=datetime.datetime.now())
                self.cfg_mgr.write_entry(section=entry, entry=self.low_tag, value=value)
        elif value > cfg_last:
            self.cfg_mgr.write_entry(section=entry, entry=self.change_amount_tag, value=value-cfg_last)
            self.cfg_mgr.write_entry(section=entry, entry=self.direction_tag, value=self.Direction.Up)
            try:
                cfg_hi = self.cfg_mgr.read_entry(section=entry, entry=self.hi_tag, default_value=self.smallest)
            except ValueError:
                cfg_hi = self.cfg_mgr.read_entry(section=entry, entry=self.hi_tag, default_value=self.smallest + .01)
            if value > cfg_hi:
                self.cfg_mgr.write_entry(section=entry, entry=self.hi_changed_tag, value=datetime.datetime.now())
                self.cfg_mgr.write_entry(section=entry, entry=self.hi_tag, value=value)
        else:
            self.cfg_mgr.write_entry(section=entry, entry=self.change_amount_tag, value=0)
            self.cfg_mgr.write_entry(section=entry, entry=self.direction_tag, value=self.Direction.NoChange)
        self.cfg_mgr.write_entry(section=entry, entry=self.last_tag, value=value)
        self.cfg_mgr.write_entry(section=entry, entry=self.last_changed_tag, value=datetime.datetime.now())
        self.cfg_mgr.write(out_file=self.file_name, stop=False, overwrite=True)
        return self.read_values(entry)


class ConfigManager:
    def __init__(self, file_name, title=None, create=False):
        self.file_name = file_name
        self.config = ConfigParser()
        self.config.optionxform = str
        self.create = create
        if not create:
            if os.path.exists(file_name):
                self.config.read(file_name)
            else:
                raise FileNotFoundError("File {} does not exist!".format(file_name))
        self.notes = []
        self.title = title
        self.values = {}

    @staticmethod
    def missing_entry(section, entry, file_name, default_value=None):
        logger = get_logger()
        logger.debug("starting")
        if default_value is None:
            log_fn = logger.critical
            message = "Required entry"
            default_value = ""
        else:
            log_fn = logger.debug
            message = "Entry"
            if default_value == "":
                default_value = "Ignoring."
            else:
                default_value = "Using default value of (" + str(default_value) + ")"
        log_fn(message + " \"" + entry + "\" in section [" + section + "] in file: " + file_name
               + " is malformed or missing.  " + str(default_value))
        if default_value == "":
            log_fn("Exiting now")
            sys.exit()

    @staticmethod
    def _insert_note(lines, line_number, note):
        if "\n" in note:
            message = note.split("\n")
        else:
            message = note
        if message is None:
            pass
        elif type(message) == str:
            lines.insert(line_number, "# " + message + ":\n")
        else:
            for l in message[:-1]:
                lines.insert(line_number, "# " + l + "\n")
                line_number += 1
            lines.insert(line_number, "# " + message[-1] + ":\n")

    def read_entry(self, section, entry, default_value, notes=None, value_type=None, use_default_if_missing=True):
        logger = get_logger()
        value = default_value
        if self.create:
            try:
                self.config.add_section(section)
            except DuplicateSectionError:
                pass
            if notes is not None:
                self.notes.append({"section": section,
                                   "entry": entry,
                                   "notes": notes})
            self.config.set(section, entry, str(default_value))
        else:
            if default_value is None:
                if value_type is None:
                    raise ConfigException("if default_value=None, value_type must be set")
                default_value = value_type
            try:
                if isinstance(default_value, str):
                    value = self.config.get(section, entry)
                elif isinstance(default_value, bool):
                    value = self.config.getboolean(section, entry)
                elif isinstance(default_value, int):
                    value = self.config.getint(section, entry)
                elif isinstance(default_value, float):
                    value = self.config.getfloat(section, entry)
                elif isinstance(default_value, dict):
                    value = ast.literal_eval(self.config.get(section, entry))
                elif isinstance(default_value, list):
                    value = ast.literal_eval(self.config.get(section, entry))
                elif isinstance(default_value, datetime.datetime):
                    value = parser.parse(self.config.get(section, entry))
                else:
                    raise ConfigException("type {} not handled for ()".format(type(default_value), default_value))
            except NoOptionError:
                logger.debug("Entry {} in section [{}] is missing.  Using default value of {}".format(entry, section,
                                                                                                      default_value))
                if not use_default_if_missing:
                    value = None
            except NoSectionError:
                logger.debug("section [{}] is missing.  Using default value of {}".format(section, default_value))
                if not use_default_if_missing:
                    value = None
        return value

    def write_entry(self, section, entry, value, note=None):
        try:
            self.config.set(section, str(entry), str(value))
        except DuplicateSectionError:
            self.config.add_section(section)
            self.config.set(section, str(entry), str(value))

        if note is not None:
            self.notes.append({"section": section,
                               "entry": entry,
                               "notes": note})

    def read_section(self, section, default_entries, notes=None):
        key_values = default_entries
        if self.create:
            try:
                self.config.add_section(section)
            except DuplicateSectionError:
                pass
            for entry in default_entries:
                self.config.set(section, str(entry), str(default_entries[entry]))
            if notes is not None:
                self.notes.append({"section": section,
                                   "entry": None,
                                   "notes": notes})
        else:
            key_values = dict()
            for (key, val) in self.config.items(section):
                key_values[key] = val
        return key_values

    def write(self, out_file, stop=True, overwrite=False):
        if os.path.isfile(out_file) and not overwrite:
            raise ConfigException("File {} exists!  You must remove it before running this".format(out_file))
        f = open(out_file, "w")
        self.config.write(f)
        f.close()
        f = open(out_file)
        lines = f.readlines()
        f.close()
        if self.title is not None:
            ConfigManager._insert_note(lines, 0, self.title)
        for note in self.notes:
            in_section = False
            line_number = 0
            for line in lines:
                if "[" + note["section"] + "]" in line:
                    if note["entry"] is None:
                        ConfigManager._insert_note(lines, line_number, note["notes"])
                        break
                    else:
                        in_section = True
                elif line.startswith("[") and line.endswith("]"):
                    in_section = False
                if in_section:
                    if line.startswith(note["entry"]):
                        ConfigManager._insert_note(lines, line_number, note["notes"])
                        break
                line_number += 1
        f = open(out_file, "w")
        contents = "".join(lines)
        f.write(contents)
        f.close()
        print("Done writing {}".format(out_file))
        if stop:
            sys.exit()


def _test_replace(filename, old_string, new_string):
    # Safely read the input filename using 'with'
    try:
        with open(filename) as f:
            s = f.read()
            if old_string not in s:
                print('"{old_string}" not found in {filename}.'.format(**locals()))
                return

        # Safely write the changed content, if found in the file
        with open(filename, 'w') as f:
            print('Changing "{old_string}" to "{new_string}" in {filename}'.format(**locals()))
            s = s.replace(old_string, new_string)
            f.write(s)
    except Exception as e:
        raise e


def _test_email_connection_config():
    file_name = "test.ini"
    for write in [True, False]:
        if write:
            if os.path.isfile(file_name):
                os.remove(file_name)
        else:
            pass
        cfg_mgr = ConfigManager(file_name,
                                "This is the title of the ini file\n"
                                "You can have multiple lines if you use line breaks", write)
        email_connection = EmailConnectionConfig(cfg_mgr)
        print(email_connection.username)
        print(email_connection.password)
        print(email_connection.from_user)
        print(email_connection.smtp_host)
        print(email_connection.port)
        if write:
            test_file = file_name
            cfg_mgr.write(test_file, stop=False)
            contents = open(test_file, "r")
            print("File contents are:")
            print("====================================================")
            print(contents.read())
            print("====================================================")
            contents.close()


def _test_config_mgr():
    file_name = "test.ini"
    for write in [True, False]:
        if write:
            if os.path.isfile(file_name):
                os.remove(file_name)
        else:
            _test_replace(file_name, "Rover", "Baily")
        cfg_mgr = ConfigManager(file_name,
                                "This is the title of the ini file\n"
                                "You can have multiple lines if you use line breaks", write)
        first = cfg_mgr.read_entry("User 1", "first name", "Joe", "This is the first name")
        last = cfg_mgr.read_entry("User 1", "last name", "Brown", "This is the last name")
        age = cfg_mgr.read_entry("User 1", "age", 12)
        is_male = cfg_mgr.read_entry("User 1", "male", True)
        weight = cfg_mgr.read_entry("User 1", "weight", 23.5)
        values = cfg_mgr.read_entry("User 1", "values", {"height": 7.5, "weight": 10, "name": "Fred"})
        weights = cfg_mgr.read_entry("User 1", "weights", [23.5, 22])
        names = cfg_mgr.read_entry("User 1", "names", ["Joe", "Fred"])
        cfg_mgr.write_entry("User 1", "male", False)
        cfg_mgr.write_entry("User 1", "parent", "Fred")
        section = cfg_mgr.read_section("user 2", {"first name": "Sally",
                                                  "last name": "Jones",
                                                  "age": 15,
                                                  "is_male": False,
                                                  "weight": 41.3},
                                       "You only get to add notes at the top of the section using this method")
        if write:
            test1 = cfg_mgr.read_entry("User 1", "dog name", "Rover")
            assert_test(test1 == "Rover", "value should be Rover")
        else:
            test1 = cfg_mgr.read_entry("User 1", "dog name", "Rover")
            assert_test(test1 == "Baily", "value should be Rover")
            test2 = cfg_mgr.read_entry("User 1", "cat name", "Tinkerbell", use_default_if_missing=False)
            assert_test(test2 is None, "missing value should be none")

        print(first)
        print(last)
        print(age)
        print(is_male)
        print(weight)
        print(values)
        print(weights)
        print(names)
        print(section)
        if write:
            test_file = file_name
            cfg_mgr.write(test_file, stop=False)
            contents = open(test_file, "r")
            print("File contents are:")
            print("====================================================")
            print(contents.read())
            print("====================================================")
            contents.close()


def _test_hi_low_vals(file_name, section, value):
    hi_low = HiLow(file_name=file_name)
    hi_value = value
    places = 10
    hi_low.write_value(entry=section, value=value)
    values = hi_low.read_values(entry=section)
    assert_test(values[HiLow.direction_tag] == HiLow.Direction.Up, "Should be moving Up")
    assert_test(values[HiLow.hi_tag] == value, "value should match")
    assert_test(values[HiLow.last_tag] == value, "value should match")
    assert_test(values[HiLow.change_amount_tag] > 1, "value should be large")

    diff = 1
    value -= diff
    hi_low.write_value(entry=section, value=value)
    values = hi_low.read_values(entry=section)
    assert_test(values[HiLow.direction_tag] == HiLow.Direction.Down, "Should be moving Down")
    assert_test(values[HiLow.low_tag] == value, "value should match")
    assert_test(values[HiLow.hi_tag] == hi_value, "value should match")
    assert_test(values[HiLow.last_tag] == value, "value should match")
    assert_test(round(values[HiLow.change_amount_tag], places) == diff, "value should be {}".format(diff))

    hi_low.write_value(entry=section, value=value)
    values = hi_low.read_values(entry=section)
    assert_test(values[HiLow.direction_tag] == HiLow.Direction.NoChange, "Should not be moving")
    assert_test(values[HiLow.low_tag] == value, "value should match")
    assert_test(values[HiLow.hi_tag] == hi_value, "value should match")
    assert_test(values[HiLow.last_tag] == value, "value should match")
    assert_test(values[HiLow.change_amount_tag] == 0, "value has not changed")

    diff = 2
    value += diff
    values = hi_low.write_value(entry=section, value=value)
    assert_test(values[HiLow.direction_tag] == HiLow.Direction.Up, "Should be moving Up")
    assert_test(values[HiLow.hi_tag] == value, "value should match")
    assert_test(values[HiLow.last_tag] == value, "value should match")
    assert_test(round(values[HiLow.change_amount_tag], places) == diff, "value should be {}".format(diff))

    diff = 1.1
    value += diff
    hi_low.write_value(entry=section, value=value)
    values = hi_low.read_values(entry=section)
    assert_test(values[HiLow.direction_tag] == HiLow.Direction.Up, "Should be moving Up")
    assert_test(values[HiLow.hi_tag] == value, "value should match")
    assert_test(values[HiLow.last_tag] == value, "value should match")
    assert_test(round(values[HiLow.change_amount_tag], places) == diff, "value should be {}".format(diff))

    diff = 10.1
    value -= diff
    hi_low.write_value(entry=section, value=value)
    values = hi_low.read_values(entry=section)
    assert_test(values[HiLow.direction_tag] == HiLow.Direction.Down, "Should be moving Down")
    assert_test(values[HiLow.low_tag] == value, "value should match")
    assert_test(values[HiLow.last_tag] == value, "value should match")
    assert_test(round(values[HiLow.change_amount_tag], places) == diff, "value should be {}".format(diff))

    value = 14
    vals = hi_low.write_value(entry=section, value=value)
    values = hi_low.read_values(entry=section)
    assert_test(vals == values, "values should match")


def _test_hi_low():
    file_name = "test.ini"
    if os.path.isfile(file_name):
        os.remove(file_name)
    value = 50
    _test_hi_low_vals(file_name, "first", value)
    value = 60
    _test_hi_low_vals(file_name, "second", value)
    value = 40
    _test_hi_low_vals(file_name, "third", value)


if __name__ == "__main__":
    log_configuration_file = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'logging.ini')
    logging.config.fileConfig(log_configuration_file)
    _test_email_connection_config()
    _test_config_mgr()
    _test_hi_low()
    print("Done!")
